import traceback
from .._utils.env import raise_fastai_import_error

import_exception = None
try:
    from ._pointcnnseg import PointCNN
    from fastai.basic_train import Learner
    from ._rand_lanet_utils import RandLANetSeg, prepare_data_dict
    from ._arcgis_model import _EmptyData
    from ._pointcnn_utils import (
        CrossEntropyPC,
        AverageMetric,
        CalculateClassificationReport,
        accuracy,
        precision,
        recall,
        f1,
    )
    import json
    from pathlib import Path
    from .._utils.common import _get_emd_path
    from .._data import _raise_fastai_import_error

    HAS_FASTAI = True
except Exception as e:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


class RandLANet(PointCNN):
    """
    Model architecture from https://arxiv.org/pdf/1911.11236v3.pdf.
    Creates RandLANet point cloud segmentation model.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    encoder_params          Optional dictionary. The keys of the dictionary are
                            `out_channels`, `sub_sampling_ratio`, `k_n`.

                              Examples:
                                {'out_channels':[16, 64, 128, 256],
                                'sub_sampling_ratio':[4, 4, 4, 4],
                                'k_n':16
                                }

                            Length of `out_channels` and `sub_sampling_ratio` should be same.
                            The length denotes the number of layers in encoder.
                              Parameter Explanation
                                - 'out_channels': Number of channels produced by each layer,
                                - 'sub_sampling_ratio': Sampling ratio of random sampling at each layer,
                                - 'k_n': Number of K-nearest neighbor for a point.

    =====================   ===========================================

    :return: `RandLANet` Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data, None)

        if not HAS_FASTAI:
            raise_fastai_import_error(
                import_exception=import_exception,
                message="This model requires module 'torch_geometric' to be installed.",
                installation_steps=" ",
            )

        self._backbone = None
        self.sample_point_num = data.max_point

        self.encoder_params = kwargs.get("encoder_params", {})
        self.encoder_params["out_channels"] = self.encoder_params.get(
            "out_channels", [16, 64, 128, 256]
        )
        self.encoder_params["num_layers"] = len(self.encoder_params["out_channels"])
        self.encoder_params["sub_sampling_ratio"] = self.encoder_params.get(
            "sub_sampling_ratio", [4] * self.encoder_params["num_layers"]
        )
        self.encoder_params["k_n"] = self.encoder_params.get("k_n", 16)
        self.encoder_params["num_classes"] = data.c
        if not isinstance(data, _EmptyData):
            data = prepare_data_dict(data, self.sample_point_num, self.encoder_params)
        self.learn = Learner(
            data,
            RandLANetSeg(self.encoder_params, data.extra_dim + 3),
            loss_func=CrossEntropyPC(data.c),
            metrics=[
                AverageMetric(accuracy),
                AverageMetric(precision),
                AverageMetric(recall),
                AverageMetric(f1),
            ],
            callback_fns=CalculateClassificationReport,
        )

        self.learn.model = self.learn.model.to(self._device)

        if pretrained_path is not None:
            self.load(pretrained_path)

    @property
    def _is_ModelInputDict(self):
        return True

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an RandLANet model object from a Deep Learning Package(DLPK)
        or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.RandLANet`  Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
        model_params = emd["ModelParameters"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping),
                chip_size=emd["ImageHeight"],
            )
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)

            ## backward compatibility.
            if not hasattr(data, "class_mapping"):
                data.class_mapping = class_mapping
                if not hasattr(data, "class2idx"):
                    data.class2idx = class_mapping
            if not hasattr(data, "color_mapping"):
                data.color_mapping = color_mapping
            if not hasattr(data, "class2idx"):
                data.class2idx = data.class_mapping
            if not hasattr(data, "classes"):
                data.classes = list(data.class2idx.values())

            data.class2idx = {int(k): int(v) for k, v in data.class2idx.items()}
            if hasattr(data, "idx2class"):
                data.idx2class = {int(k): int(v) for k, v in data.idx2class.items()}

            data.color_mapping = {int(k): v for k, v in data.color_mapping.items()}

            ## Below are the lines to make save function work
            data.chip_size = None
            data._image_space_used = None
            data.dataset_type = "PointCloud"

        return cls(data, **model_params, pretrained_path=str(model_file))



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="EncoderDecoder",
    pretrained="open-mmlab://msra/hrnetv2_w48",
    backbone=dict(
        type="HRNet",
        norm_cfg=norm_cfg,
        norm_eval=False,
        extra=dict(
            stage1=dict(
                num_modules=1,
                num_branches=1,
                block="BOTTLENECK",
                num_blocks=(4,),
                num_channels=(64,),
            ),
            stage2=dict(
                num_modules=1,
                num_branches=2,
                block="BASIC",
                num_blocks=(4, 4),
                num_channels=(48, 96),
            ),
            stage3=dict(
                num_modules=4,
                num_branches=3,
                block="BASIC",
                num_blocks=(4, 4, 4),
                num_channels=(48, 96, 192),
            ),
            stage4=dict(
                num_modules=3,
                num_branches=4,
                block="BASIC",
                num_blocks=(4, 4, 4, 4),
                num_channels=(48, 96, 192, 384),
            ),
        ),
    ),
    decode_head=dict(
        type="FCNHead",
        in_channels=[48, 96, 192, 384],
        in_index=(0, 1, 2, 3),
        channels=sum([48, 96, 192, 384]),
        input_transform="resize_concat",
        kernel_size=1,
        num_convs=1,
        concat_input=False,
        dropout_ratio=-1,
        num_classes=19,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/hrnet/fcn_hr48_512x1024_160k_cityscapes/fcn_hr48_512x1024_160k_cityscapes_20200602_190946-59b7973e.pth"



from arcgis import GIS
from typing import Optional, Dict, Union, List
from ._task import Task
from ._util import _Util


class Feed(Task):
    """
    The ``Feed`` class implements Task and provides public facing methods to access Feed API endpoints.
    """

    _id = ""
    _gis = None
    _util = None
    _item = None

    def __init__(self, gis: GIS, util: _Util, item: Optional[Dict] = None):
        self._gis = gis
        self._util = util
        self._item = item
        self._id = item["id"]

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "<%s id:%s label:%s>" % (
            type(self).__name__,
            self._id,
            self._item["label"],
        )

    # ----------------------------------------------------------------------
    def start(self) -> Dict:
        """
        Start the Feed for the given ID.

        :return: response of feed start

        .. code-block:: python

            # start feed

            # Method: <item>.start

            sample_feed.start()

        """
        return self._util._start("feed", self._id)

    # ----------------------------------------------------------------------
    def stop(self) -> Dict:
        """
        Stop the Feed for the given ID.
        Return True if the Feed was successfully stopped.

        :return: boolean

        .. code-block:: python

            # stop feed

            # Method: <item>.stop

            sample_feed.stop()

        """
        return self._util._stop("feed", self._id)

    # ----------------------------------------------------------------------
    @property
    def status(self) -> Dict:
        """
        Get the status of the running Feed for the given ID.
        :return: response of Feed status

        .. code-block:: python

            # Retrieve status of sample_feed

            # Method: <item>.status

            Sample_feed = feeds.get("id")
            status = sample_feed.status
            status
        """
        return self._util._status("feed", self._id)

    # ----------------------------------------------------------------------
    @property
    def metrics(self) -> Dict:
        """
        Get the metrics of the running Feed for the given ID.

        :return: response of feed metrics

        .. code-block:: python

            # Method: <item>.metrics

            # Retrieve metrics of sample_feed

            metrics = sample_feed.metrics
            metrics
        """
        return self._util._metrics("feed/metrics", self._id)

    # ----------------------------------------------------------------------
    def delete(self) -> bool:
        """
        Deletes an existing Feed instance.

        :return: A boolean containing True (for success) or
         False (for failure) a dictionary with details is returned.

        .. code-block:: python

            # Delete a feed

            # Method: <item>.delete()

            sample_feed.delete()
        """
        return self._util._delete("feed", self._id)



from ._contentds import (
    ItemTypeEnum,
    ItemProperties,
    CreateServiceParameter,
    MetadataFormatEnum,
    ServiceTypeEnum,
)
from ._viewdc import ViewLayerDefParameter
from ._sfilters import SpatialFilter, SpatialRelationship

__all__ = [
    "ItemTypeEnum",
    "ItemProperties",
    "CreateServiceParameter",
    "MetadataFormatEnum",
    "ServiceTypeEnum",
    "ViewLayerDefParameter",
    "SpatialFilter",
    "SpatialRelationship",
]



import os
import torch
import json
import traceback

HAS_TRANSFORMER = True

try:
    import torch
    from transformers import AutoModelForSequenceClassification
    from ._arcgis_transformer import ArcGISTransformer, transformer_seq_length
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False
    transformer_seq_length = 512

    class ArcGISTransformer:
        pass


backbone_models_map = {
    "bert": (
        "bert-base-cased",
        "bert-base-uncased",
        "bert-large-cased",
        "bert-large-uncased",
    ),
    "albert": (
        "albert-base-v1",
        "albert-base-v2",
        "albert-large-v1",
        "albert-large-v2",
        "albert-xlarge-v1",
        "albert-xlarge-v2",
        "albert-xxlarge-v1",
        "albert-xxlarge-v2",
    ),
    "roberta": ("roberta-base", "roberta-large", "distilroberta-base"),
    "distilbert": ("distilbert-base-cased", "distilbert-base-uncased"),
    "xlnet": ("xlnet-base-cased", "xlnet-large-cased"),
    "xlm": (
        "xlm-mlm-ende-1024",
        "xlm-mlm-enfr-1024",
        "xlm-mlm-xnli15-1024",
        "xlm-mlm-en-2048",
    ),
    "flaubert": (
        "flaubert/flaubert_small_cased",
        "flaubert/flaubert_base_cased",
        "flaubert/flaubert_base_uncased",
        "flaubert/flaubert_large_cased",
    ),
    "xlm-roberta": ("xlm-roberta-base", "xlm-roberta-large"),
    "longformer": ("allenai/longformer-base-4096",),
    "mobilebert": ("google/mobilebert-uncased",),
    "electra": ("google/electra-base-discriminator", "google/electra-base-generator"),
    "bart": ("facebook/bart-base", "facebook/bart-large"),
    "camembert": ("camembert-base",),
    "funnel": (
        "funnel-transformer/small",
        "funnel-transformer/small-base",
        "funnel-transformer/medium",
        "funnel-transformer/medium-base",
    ),
}

transformer_architectures = [
    "BERT",
    "RoBERTa",
    "DistilBERT",
    "ALBERT",
    "FlauBERT",
    "CamemBERT",
    "XLNet",
    "XLM",
    "XLM-RoBERTa",
    "Bart",
    "ELECTRA",
    "Longformer",
    "MobileBERT",
    "Funnel",
]

backbone_models_reverse_map = {
    x: key for key, val in backbone_models_map.items() for x in val
}


class TransformerForTextClassification(ArcGISTransformer):
    _supported_backbones = transformer_architectures

    def __init__(
        self,
        architecture,
        pretrained_model_name,
        config=None,
        pretrained_model_path=None,
        seq_len=transformer_seq_length,
    ):
        super().__init__(
            architecture,
            pretrained_model_name,
            config,
            pretrained_model_path,
            task="classification",
        )
        self._seq_len = seq_len
        self._max_seq_len = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % type(self).__name__

    def _process_config(self):
        """
        Process the model config to convert the class values to integers in the
        id2label attribute (which is a dictionary) of the model config
        """
        self._config.id2label = {int(x): y for x, y in self._config.id2label.items()}
        """
        The `funnel` transformer has 2 type of attention `relative_shift` & `factorized`. The transformer 
        config uses `relative_shift` attention type by default. When tested the text classifier model 
        on twitter dataset, we were getting errors while training the model. Seems like a bug in the 
        library. Hence using the `factorized` attention type here.
        """
        if "funnel" in self._transformer_pretrained_model_name:
            self._config.attention_type = "factorized"

    def _set_max_seq_length(self):
        """
        Set the max-seq-len parameter to avoid out of memory issues for transformers
        like Bart, Longformer, XLNet
        """
        self._max_seq_len = min(self._tokenizer.model_max_length, self._seq_len)

    @classmethod
    def _available_backbone_models(cls, architecture):
        """
        Provides a list of available models for a given transformer architecture

        =====================   =================================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------------
        architecture            Required string. The transformer architecture for
                                which we wish to get the available models
        ---------------------   -------------------------------------------------
        return: tuple containing available models for the given architecture
        """
        if architecture.lower() in backbone_models_map:
            return backbone_models_map[architecture.lower()]
        else:
            return (
                f"Error, wrong architecture name - {architecture} supplied. "
                f"Please choose from {cls._supported_backbones}"
            )

    def save(self, model_path):
        """
        Method to save the fine-tuned model to the disk

        =====================   =================================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------------
        model_path              Required string. The disk location where the
                                fine-tuned model has to be saved
        ---------------------   -------------------------------------------------
        """
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        model_params = {
            "architecture": self._transformer_architecture,
            "pretrained_model": self._transformer_pretrained_model_name,
        }
        file_path = os.path.join(model_path, self._outfile)
        out_file = open(file_path, "w")
        json.dump(model_params, out_file)
        self._transformer.save_pretrained(model_path)
        # self._tokenizer.save_pretrained(model_path)

    @classmethod
    def load_pretrained_model(cls, path):
        """
        Method to load the fine-tuned model which was saved on the disk

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The disk location from
                                where the fine-tuned model has to be loaded
        ---------------------   -------------------------------------------
        """
        architecture_file_path = os.path.join(path, cls._outfile)
        if os.path.exists(architecture_file_path):
            with open(architecture_file_path) as f:
                data = json.load(f)
            architecture = data.get("architecture")
            pretrained_model_name = data.get("pretrained_model")
            obj = TransformerForTextClassification(
                architecture, pretrained_model_name, pretrained_model_path=path
            )
            obj.init_model()
            return obj
        else:
            raise Exception(f"{cls._outfile} not present at {path}")

    def forward(self, input_ids, **kwargs):
        """
        Return only the logits from the transfomer model

        =====================   ==============================================
        **Parameter**            **Description**
        ---------------------   ----------------------------------------------
        input_ids               tensor object. tensor containing the token-ids
                                got by calling tokenizer.encode method to the
                                input text. works same for mini-batches
        ---------------------   ----------------------------------------------

        :return: the logits from the transformer model
        """
        logits = self._transformer(input_ids)[0]
        return logits

    def _load_transformer(self):
        """
        Method to load the transformer model. For this class we want to load the
        transformer model for performing Sequence Classification tasks
        """
        if self._pretrained_model_path and os.path.exists(self._pretrained_model_path):
            base_path, extension = os.path.splitext(self._pretrained_model_path)
            path = base_path + ".pth"
            if extension == ".emd" and os.path.exists(path):
                self._transformer = AutoModelForSequenceClassification.from_pretrained(
                    path, config=self._config
                )
            else:
                self._transformer = AutoModelForSequenceClassification.from_pretrained(
                    self._pretrained_model_path
                )
        else:
            self._transformer = AutoModelForSequenceClassification.from_pretrained(
                self._transformer_pretrained_model_name, config=self._config
            )

    def predict_class(self, text, device, is_multilabel_problem=False, thresh=None):
        """
        Method to predict the class labels for an input text

        =====================   ==============================================
        **Parameter**            **Description**
        ---------------------   ----------------------------------------------
        text                    Required string. The text for which we wish to
                                predict the class
        ---------------------   ----------------------------------------------

        :return: the predicted label and the corresponding confidence score.
        """
        # device_type = next(self._transformer.parameters()).device.type
        # device = torch.device("cuda:0" if device_type == "cuda" else "cpu")
        sequence = torch.tensor(
            [
                self._tokenizer.encode(
                    text, max_length=self._max_seq_len, truncation=True
                )
            ]
        ).to(device)
        logits = self._transformer(sequence)[0]
        if is_multilabel_problem:
            results = torch.sigmoid(logits)[0]
            category, raw_pred = [], []
            for idx, x in enumerate(results):
                pred = 0
                if x.item() > thresh:
                    pred = 1
                    category.append(self._config.id2label[idx])
                raw_pred.append(pred)
            return ";".join(category), raw_pred, [round(x, 4) for x in results.tolist()]
        else:
            results = torch.softmax(logits, dim=1)[0]
            clas = self._config.id2label[torch.argmax(logits, dim=1)[0].item()]
            score = results[torch.argmax(logits, dim=1)[0]]
            return clas, score.item()

    def predict_class_batch(
        self, text_batch, device, is_multilabel_problem=False, thresh=None
    ):
        """
        Method to predict the class labels for an input text

        =====================   ==============================================
        **Parameter**            **Description**
        ---------------------   ----------------------------------------------
        text_batch              Required list. The list of text for which we wish to
                                predict the class
        ---------------------   ----------------------------------------------

        :return: the predicted label and the corresponding confidence score.
        """
        # device_type = next(self._transformer.parameters()).device.type
        # device = torch.device("cuda:0" if device_type == "cuda" else "cpu")
        encodings = self._tokenizer.batch_encode_plus(
            text_batch, max_length=self._max_seq_len, padding=True, truncation=True
        )
        batch_token_ids = encodings["input_ids"]
        results_list = []
        sequence = torch.tensor(batch_token_ids).to(device)
        with torch.no_grad():
            logits = self._transformer(sequence)[0]

        if is_multilabel_problem:
            results = torch.sigmoid(logits)
            for res in results:
                category, raw_pred = [], []
                for idx, x in enumerate(res):
                    pred = 0
                    if x.item() > thresh:
                        pred = 1
                        category.append(self._config.id2label[idx])
                    raw_pred.append(pred)
                results_list.append(
                    (";".join(category), raw_pred, [round(x, 4) for x in res.tolist()])
                )
        else:
            results = torch.softmax(logits, dim=1).cpu()
            res_class = [
                self._config.id2label[i] for i in torch.argmax(results, dim=1).numpy()
            ]
            res_score = torch.max(results, dim=1).values.numpy()
            return list(zip(res_class, res_score))

        return results_list



import traceback
from ._arcgis_model import ArcGISModel, _EmptyData
from .._data import _raise_fastai_import_error
import warnings

import_exception = None

try:
    import gc
    from fastai.basic_train import Learner
    import torch
    import numpy as np
    from .._utils.pointcloud_od import plot_results, confusion_matrix3d, predict_h5
    from .._utils.pointcloud_data import raise_maxpoint_warning
    from fastprogress.fastprogress import progress_bar
    from torch import LongTensor
    from ._ssd_utils import compute_ap_score, AveragePrecision
    import json
    from pathlib import Path
    from .._utils.common import _get_emd_path

    HAS_FASTAI = True
except Exception as e:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


def mmdet3d_model(data, **kwargs):
    from ._mmdet3d_utils import get_model

    return get_model(data, **kwargs)


def mmdet3d_loss(model_output, *model_target):
    return model_output[1]["loss"]


class MMDetection3D(ArcGISModel):
    """
    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    data                            Required fastai Databunch. Returned data object from
                                    :meth:`~arcgis.learn.prepare_data` function.
    -----------------------------   ---------------------------------------------
    model                           Required model name or path to the configuration file
                                    from :class:`~arcgis.learn.MMDetection3D` repository.
                                    The list of the supported models can be queried using
                                    :attr:`~arcgis.learn.MMDetection.supported_models`.
    -----------------------------   ---------------------------------------------
    pretrained_path                 Optional string. Path where pre-trained model is
                                    saved.
    =============================   =============================================

     **kwargs**

    =============================   =============================================
    **Parameter**                   **Description**
    -----------------------------   ---------------------------------------------
    voxel_parms                     Optional dictionary. The keys of the dictionary are
                                    `voxel_size`, `voxel_points`, and `max_voxels`. The
                                    default value of `voxel_size`,`voxel_points`, and
                                    `max_voxels` are automatically calculated based on
                                    the 'block size', 'object size' and 'average no.
                                    of points per block' of the exported data.

                                    Example:
                                        |    {'voxel_size': [0.05, 0.05, 0.1],
                                        |    'voxel_points': 10,
                                        |    'max_voxels':[20000, 40000],
                                        |    }

                                    Parameter Explanation:

                                    - 'voxel_size': List of voxel dimensions in meter
                                      [x,y,z],
                                    - 'voxel_points': An Int, that decides the maximum
                                      number of points per voxel,
                                    - 'max_voxels': List of maximum number of voxels in
                                      [training, validation].

                                    Default: None.
    =============================   =============================================

    :return: :class:`~arcgis.learn.MMDetection3D` Object
    """

    def __init__(self, data, model="SECOND", pretrained_path=None, **kwargs):
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        kwargs["voxel_parms"] = kwargs.get("voxel_parms", {})
        self._kwargs = kwargs
        self._kwargs["model"] = model
        self._check_dataset_support(data)
        super().__init__(data, None, **kwargs)
        self._backbone = None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model, self._config = mmdet3d_model(data, **kwargs)
        self.learn = Learner(data, model, loss_func=mmdet3d_loss)
        self.learn.metrics = [AveragePrecision(self, data.c, mode_3d=True)]
        self._reset_thresh()
        if pretrained_path is not None:
            self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    supported_models = ["SECOND"]
    """
    List of models supported by this class.
    """

    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder. Helps in choosing the
        optimum learning rate for training the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        allow_plot              Optional boolean. Display the plot of losses
                                against the learning rates and mark the optimal
                                value of the learning rate on the plot.
                                The default value is 'True'.
        =====================   ===========================================
        """
        lr = super().lr_find(allow_plot)
        lr = min(max(lr, 5e-05), 3e-03)
        return lr

    def _free_memory(self):
        gc.collect()
        torch.cuda.empty_cache()

    def _reset_thresh(self, detect_thresh=0.2, nms_overlap=0.2):
        self.learn.model.bbox_head.test_cfg.score_thr = detect_thresh
        self.learn.model.bbox_head.test_cfg.nms_thr = nms_overlap
        self._config.model.test_cfg.score_thr = detect_thresh
        self._config.model.test_cfg.nms_thr = nms_overlap

    def _pred_batch(self, data, detect_thresh=0.2, nms_overlap=0.5):
        self.learn.model.bbox_head.test_cfg.score_thr = detect_thresh
        self.learn.model.bbox_head.test_cfg.nms_thr = nms_overlap
        self.learn.model.eval()
        self.learn.model.prediction = True
        with torch.no_grad():
            pred = self.learn.model(data)
        self.learn.model.prediction = False
        return pred

    def show_results(self, rows=2, detect_thresh=0.3, nms_overlap=0.01, **kwargs):
        """
        Displays the results of the trained model on a part of validation/train set.
        Colors of the PointCloud are only used for better visualization, and it does
        not depict the actual classcode colors. Visualization of data, exported in a
        geographic coordinate system is not yet supported.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        color_mapping           Optional dictionary. Mapping from object id
                                to RGB values. Colors of the PointCloud via
                                color_mapping are only used for better
                                visualization, and it does not depict the
                                actual classcode colors. Default value example:
                                {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
        ---------------------   -------------------------------------------
        max_display_point       Optional integer. Maximum number of points
                                to display. Default is 20000. A warning will
                                be raised if the total points to display exceeds
                                this parameter. Setting this parameter will
                                randomly sample the specified number of points
                                and once set, it will be used for future uses.
        ---------------------   -------------------------------------------
        view_type               Optional string. Dataset type to display the
                                results.
                                    * ``valid`` - For validation set.
                                    * ``train`` - For training set.
                                Default: 'valid'.
        =====================   ===========================================
        """
        ds_type = kwargs.get("view_type", "valid")
        max_display_point = kwargs.get("max_display_point", 20000)
        save_html = kwargs.get("save_html", False)
        kwargs["color_mapping"] = kwargs.get("color_mapping", self._data.color_mapping)
        kwargs["class_mapping"] = self._data.class_mapping
        batch_size = self._data.valid_dl.batch_size
        row_idx = 0
        while row_idx < rows:
            if ds_type == "valid":
                data = self._data.valid_ds.get_batch(batch_size)
            else:
                data = self._data.train_ds.get_batch(batch_size)
            preds = self._pred_batch(data, detect_thresh, nms_overlap)

            for idx in range(batch_size):
                points = data["points"][idx][:, :3]
                if points.shape[0] > max_display_point:
                    raise_maxpoint_warning(
                        row_idx, kwargs, None, max_display_point, save_html
                    )
                    mask = torch.from_numpy(
                        np.random.randint(
                            0, points.shape[0], max_display_point, dtype=np.int64
                        )
                    )
                else:
                    mask = torch.arange(0, points.shape[0])
                points = points[mask]
                gt_boxes = data["gt_bboxes_3d"][idx]
                gt_boxes_labels = data["gt_labels_3d"][idx].detach().cpu()
                gt_points_labels = gt_boxes.points_in_boxes_part(points).detach().cpu()
                pred_boxes = preds[idx]["boxes_3d"]
                pred_boxes_labels = preds[idx]["labels_3d"]
                pred_points_labels = (
                    pred_boxes.points_in_boxes_part(points).detach().cpu()
                )
                points = points.detach().cpu()
                plot_results(
                    row_idx,
                    points,
                    gt_boxes,
                    gt_boxes_labels,
                    gt_points_labels,
                    pred_boxes,
                    pred_boxes_labels,
                    pred_points_labels,
                    idx2class=self._data.idx2class,
                    **kwargs,
                )
                row_idx += 1
                if row_idx == rows:
                    break
        self._free_memory()
        self._reset_thresh()

    def average_precision_score(
        self, detect_thresh=0.3, iou_thresh=0.1, nms_overlap=0.01, mean=False, **kwargs
    ):
        """
        Computes average precision on the validation/train set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
                                Default: 0.3.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
                                Default: 0.1.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
                                Default: 0.01.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
                                Default: False.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        view_type               Optional string. Dataset type to display the
                                results.
                                    * ``valid`` - For validation set.
                                    * ``train`` - For training set.
                                Default: 'valid'.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        ds_type = kwargs.get("view_type", "valid")
        show_progress = kwargs.get("show_progress", True)
        if ds_type == "valid":
            dl = self._data.valid_dl
        else:
            dl = self._data.train_dl
        classes, n_gts = (
            LongTensor(range(self._data.c)),
            torch.zeros(self._data.c).long(),
        )
        tps, pred_scores, pred_clas = [], [], []
        with torch.no_grad():
            for input, target in progress_bar(dl, display=show_progress):
                pred = self._pred_batch(input, detect_thresh, nms_overlap)
                batch_tps, batch_score, batch_clas, n_gts = confusion_matrix3d(
                    pred, target, n_gts, classes, iou_thresh
                )
                tps.extend(batch_tps)
                pred_scores.extend(batch_score)
                pred_clas.extend(batch_clas)
        aps = compute_ap_score(
            tps, pred_scores, pred_clas, n_gts, self._data.c, mode_3d=True
        )
        self._free_memory()
        self._reset_thresh()
        if mean:
            return np.mean(aps)
        else:
            return dict(zip(self._data.classes, aps))

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "average_precision"]

    @property
    def _model_metrics(self):
        return {"average_precision_score": self._get_model_metrics()}

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return self.average_precision_score()

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            model_accuracy = self.learn.recorder.metrics[
                self.learn._best_epoch  # index using best epoch.
            ][0]

        return float(model_accuracy)

    @property
    def _is_mmsegdet(self):
        return True

    @property
    def _is_mm3d(self):
        return True

    def _get_emd_params(self, save_inference_file):
        emd_template = {"DataAttributes": {}, "ModelParameters": {}}
        emd_template["ModelType"] = "PointCloudDetection"
        emd_template["ModelParameters"]["kwargs"] = self._kwargs
        emd_template["DataAttributes"]["block_size"] = self._data.block_size
        emd_template["DataAttributes"]["max_point"] = self._data.max_point
        emd_template["DataAttributes"]["extra_features"] = self._data.extra_features
        emd_template["DataAttributes"]["pc_type"] = self._data.pc_type
        emd_template["DataAttributes"]["scale_factor"] = self._data.scale_factor
        emd_template["DataAttributes"]["color_mapping"] = self._data.color_mapping
        emd_template["DataAttributes"]["class_mapping"] = self._data.class_mapping
        emd_template["DataAttributes"]["classes"] = self._data.classes
        emd_template["DataAttributes"]["class2idx"] = self._data.class2idx
        emd_template["DataAttributes"]["idx2class"] = self._data.idx2class
        emd_template["DataAttributes"]["remap"] = self._data.remap
        emd_template["DataAttributes"]["z_range"] = self._data.z_range
        emd_template["DataAttributes"]["range"] = self._data.range
        emd_template["DataAttributes"]["average_box_size"] = self._data.average_box_size
        emd_template["DataAttributes"]["anchor_range"] = self._data.anchor_range
        emd_template["DataAttributes"]["num_features"] = self._data.num_features
        emd_template["DataAttributes"]["features_to_keep"] = self._data.features_to_keep
        emd_template["DataAttributes"]["voxel_size"] = self._data.voxel_size
        emd_template["DataAttributes"][
            "no_of_points_per_tile"
        ] = self._data.no_of_points_per_tile

        emd_template["Classes"] = []
        class_data = {}
        for class_name in self._data.classes:
            class_data["Value"] = class_name
            class_data["Name"] = self._data.class_mapping[class_name]
            class_data["Color"] = self._data.color_mapping[class_name]
            emd_template["Classes"].append(class_data.copy())

        if hasattr(self._data, "statistics") and self._data.statistics[
            "parameters"
        ].get("excludedClasses", False):
            emd_template["excludedClasses"] = self._data.statistics["parameters"][
                "excludedClasses"
            ]

        if hasattr(self._data, "statistics") and self._data.statistics.get(
            "blockShape", False
        ):
            emd_template["blockShape"] = self._data.statistics.get("blockShape")

        return emd_template

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.MMDetection` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.MMDetection3D` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
        color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping),
                chip_size=emd["ImageHeight"],
            )
            data.emd_path = emd_path
            data.emd = emd
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping

        return cls(
            data, **emd["ModelParameters"]["kwargs"], pretrained_path=str(model_file)
        )

    def predict_h5(self, path, output_path=None, **kwargs):
        """
        This method is used for infrencing using HDF file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The path to folder where the HDF
                                files which needs to be predicted are present.
        ---------------------   -------------------------------------------
        output_path             Optional string. The path to folder where to dump
                                the resulting HDF files. Defaults to `results`
                                folder in input path.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        batch_size              Optional integer. The number of blocks to process
                                in one batch. Default is set to 1.
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered valid.
                                Default: 0.1.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
                                Default: 0.6.
        =====================   ===========================================

        :return: Path where files are dumped.
        """
        return predict_h5(self, path, output_path, **kwargs)

    def unfreeze(self):
        """
        Not implemented for this model as
        none of the layers are frozen by default.
        """



""" Defines Exception subclasses for errors raised by the survey123-python-api. """


class SurveyError(Exception):
    """
    Abstract base class for exceptions thrown by the Survey123-python-api
    """

    def __init__(self, message):
        """
        :param message: A human readable message describing the error.
        """

        super().__init__(self)
        self.message = message

    def __str__(self):
        return self.message


class ServerError(SurveyError):
    """
    An error that was generated by a service in the backend, such as a portal, feature service, or API.
    """

    def __init__(self, error):
        """:param error: An error string returned by the server."""
        message = "An error occurred while generating reports. Details: {}".format(
            error
        )
        super().__init__(message)
        self.errors = error

    def __repr__(self):
        return "ServerError({})".format(repr(self.errors))



class InvitationManager(object):
    """

    The ``InvitationManager`` provides functionality to see the existing invitations
    set out via email to your organization.

    .. note::
        The manager has the ability to delete
        any invitation sent out by an organization.

    """

    _gis = None
    _url = None
    _invites = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis):
        self._url = url
        self._gis = gis

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< InvitationManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< InvitationManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __len__(self):
        url = self._url
        params = {"f": "json", "num": 100, "start": 1}
        res = self._gis._con.get(url, params)
        return res["total"]

    # ----------------------------------------------------------------------
    def list(self):
        """
        The ``list`` method retrieves all the organization's invitations.

        :return:
            A List of the organization's invitations

        """
        invites = []
        from arcgis.gis import GIS

        isinstance(self._gis, GIS)
        url = self._url
        params = {"f": "json", "num": 100, "start": 1}
        res = self._gis._con.get(url, params)
        invites = res["invitations"]
        while res["nextStart"] > -1:
            params["start"] += 100
            res = self._gis._con.get(url, params)
            invites.extend(res["invitations"])
        return invites

    def get(self, invite_id: str):
        """
        The ``get`` method retrieves information about a single invitation.

        :return:
            A dictionary

        """
        url = self._url + "/{id}".format(id=invite_id)
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    # ----------------------------------------------------------------------
    def _accept(self, invite_id):
        """Accepts the Invitation"""
        url = f"{self._url}/{invite_id}/accept"
        return self._gis._con.post(url, params={"f": "json"})

    # ----------------------------------------------------------------------
    def _decline(self, invite_id):
        """Reject the Invitation"""
        url = f"{self._url}/{invite_id}/decline"
        return self._gis._con.post(url, params={"f": "json"})

    # ----------------------------------------------------------------------
    def manage_invitations(self, accepts: list = None, declines: list = None) -> dict:
        """
        The ``manage_invitations`` method allows users to Accept/Decline invitations by providing a list of
        invitation IDs.

        :return:
            A List of invitation IDs
        """
        results = []
        if accepts and isinstance(accepts, (list, tuple)):
            accepts = [self._accept(invid) for invid in accepts]
        if declines and isinstance(declines, (list, tuple)):
            declines = [self._decline(invid) for invid in declines]
        return {"accepts": accepts, "declines": declines}

    # ----------------------------------------------------------------------
    def delete(self, invite_id: str):
        """
        The ``delete`` method deletes an invitation by ID

        :return:
            A boolean indicating success (True), or failure (False)
        """
        url = self._url + "/{id}/delete".format(id=invite_id)
        params = {"f": "json"}
        res = self._gis._con.post(url, params)
        if "success" in res:
            return res["success"]
        return res



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS


###########################################################################
class LicenseManager(_BaseKube):
    """The license manager for the Kubernetes deployment."""

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    def load(self, license_file: str, overwrite: bool) -> bool:
        """
        `load` imports and applies an ArcGIS Server authorization file. By
        default, this operation will append authorizations from the
        imported license file to the current authorizations. Optionally,
        you can select the overwrite option to fully replace the current
        authorizations with those from the imported license file.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        license_file           Required string. The ArcGIS Server authorization file (either in
                               .epc or .prvc file format).
        ------------------     --------------------------------------------------------------------
        overwrite              Required bool. Specifies whether the authorizations in the imported
                               license file will fully replace or be appended to the current
                               authorizations. If true, the authorizations from the imported
                               license file will replace the current authorizations. If false, the
                               authorizations from the imported license file will be appended to
                               the current authorizations.
        ==================     ====================================================================

        :return: Boolean. True if successful else False.

        """
        url = f"{self._url}/validateLicense"
        params = {
            "f": "json",
            "overwrite": overwrite,
        }
        files = file = {"licenseFile": license_file}
        res = self._con.post(url, params, files=file)
        return res.get("status", "failed") == "success"



from __future__ import annotations
from typing import Optional
from urllib.request import HTTPError
from arcgis._impl.common._isd import InsensitiveDict
from ._base import _BaseKube
from ._logs import LogManager
from ._datastores import DataStores
from ._overview import Overview
from ._usage import UsageStatistics
from ._mode import Mode
from ._system import SystemManager
from ._jobs import JobManager
from arcgis.gis.admin._license import LicenseManager
from arcgis.gis import Item, User
from arcgis.apps.tracker._location_tracking import LocationTrackingManager


class KubernetesAdmin(_BaseKube):
    """
    Kubernetes Administration Class. This class is not meant to be initialized
    directly, but instead is returned by the _admin_ property on the
    :class:`gis <arcgis.gis.GIS>` object when logged in as an administrator.

    .. code-block:: python

        #Usage Example:

        >>> gis = GIS(profile="your_kubernetes_admin_profile")
        >>> kube_admin = gis.admin
        >>> kube_admin

        <KubernetesAdmin at https://kubenetes.example.com/arcgis/admin>
    """

    _url = None
    _gis = None
    _con = None
    _log = None
    _ds = None
    _sm = None
    _sp = None
    _mode = None
    _catalog = None
    _idp = None
    _whm = None
    _security = None
    _services = None
    _license = None
    _metadata = None
    _uploads = None
    _properties = None
    _organizations = None
    _category_schema = None
    _jobs = None
    _collaborations = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis):
        """class initializer"""
        super(KubernetesAdmin, self)
        self._url = url
        self._gis = gis
        self._con = gis._con
        self._init(gis._con)

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = InsensitiveDict(result)
            else:
                self._json_dict = {}
                self._properties = InsensitiveDict({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = InsensitiveDict({})

    # ----------------------------------------------------------------------
    @property
    def overview(self) -> Overview:
        """
        Provides access to the :class:`~arcgis.gis.kubernetes.overview`
        resource to access persisted cache or real-time information.

        :return: :class:`~arcgis.gis.kubernetes.Overview` object

        """
        url = f"{self._url}/overview"
        return Overview(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def usage(self) -> UsageStatistics:
        """
        Provides access to the metrics viewer and metrics API tools.

        :return:
            An :class:`~arcgis.gis.kubernetes.UsageStatistics` object.

        """
        url = f"{self._url}/usagestatistics"
        return UsageStatistics(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def logs(self) -> LogManager:
        """Accesses a :class:`~arcgis.gis.kubernetes.LogManager` object to
        manage and query the Kubernetes logs."""
        if self._log is None:
            url = f"{self._url}/logs"
            self._log = LogManager(url, gis=self._gis)
        return self._log

    # ----------------------------------------------------------------------
    @property
    def mode(self) -> Mode:
        """Provides access to a :class:`~arcgis.gis.kubernetes.Mode` object
        to help manage `service deployment modes <https://enterprise-k8s.arcgis.com/en/latest/administer/service-modes.htm>`_.
        """
        if self._mode is None:
            self._mode = Mode(url=f"{self._url}/mode", gis=self._gis)
        return self._mode

    # ----------------------------------------------------------------------
    @property
    def datastores(self) -> DataStores:
        """
        Provides access to the _Datastore Manager_, allowing the administrator
        to manage registered datastores.

        :return: A Kubernetes :class:`~arcgis.gis.kubernetes.DataStores` object.
        """
        if self._ds is None:
            url = self._url + "/data"
            self._ds = DataStores(url=url, gis=self._gis)
        return self._ds

    # ----------------------------------------------------------------------
    @property
    def system(self) -> SystemManager:
        """
        This is a collection of system-wide resources for your deployment
        such as the configuration store, licenses, and deployment-wide
        security.

        :return:
            :class:`~arcgis.gis.kubernetes.SystemManager` object.

        """
        if self._sm is None:
            url = self._url + "/system"
            self._sm = SystemManager(url=url, gis=self._gis)
        return self._sm

    # ----------------------------------------------------------------------
    @property
    def jobs(self) -> JobManager:
        """
        This resource is a collection of the asynchronous
        :class:`jobs <arcgis.gis.kubernetes.Job>` created in your
        deployment. When operations that support asynchronous executions are
        run with the async option enabled, a new job entry is created that
        can be queried for its current status and messages.
        """
        if self._jobs is None:
            url = self._url + "/jobs"
            self._jobs = JobManager(url=url, gis=self._gis)
        return self._jobs

    # ----------------------------------------------------------------------
    @property
    def license(self) -> LicenseManager:
        """
        Provides access to the :class:`~arcgis.gis.kubernetes.LicenseManager`,
        and its set of tools to access and manage user licenses and
        entitlements.
        """
        if self._license is None:
            url = self._gis._portal.resturl + "portals/self/purchases"
            self._license = LicenseManager(url=url, gis=self._gis)
        return self._license

    # ----------------------------------------------------------------------
    @property
    def category_schema(self):
        """This resource allows for the setting and manipulating of catagory schemas.

        :returns:
            :class:`~arcgis.gis.admin.CategoryManager` object
        """
        if self._category_schema is None:
            from arcgis.gis.admin._catagoryschema import CategoryManager

            self._category_schema = CategoryManager(gis=self._gis)
        return self._category_schema

    # ----------------------------------------------------------------------
    @property
    def _idp(self):
        """
        This resource allows for the setting and configuration of the identity provider
        """
        if self._idp is None:
            from arcgis.gis.admin._idp import IdentityProviderManager

            self._idp = IdentityProviderManager(gis=self._gis)
        return self._idp

    # ----------------------------------------------------------------------
    def scheduled_tasks(
        self,
        item: Optional[Item] = None,
        active: Optional[bool] = None,
        user: Optional[User] = None,
        types: Optional[str] = None,
    ):
        """
        This method allows organization admins to see all scheduled tasks
        on the organization.

        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        item              Optional Item. The item to query tasks about.
        ----------------  -------------------------------------------------------------------------------
        active            Optional Bool. Queries tasks based on active status.
        ----------------  -------------------------------------------------------------------------------
        user              Optional User. Search for tasks for a single user.
        ----------------  -------------------------------------------------------------------------------
        types             Optional String. The type of notebook execution for the item:

                          * `ExecuteNotebook`
                          * `UpdateInsightsWorkbook`
        ================  ===============================================================================


        :return: List of :class:`Tasks <arcgis.gis.tasks.Task>`.

        """
        _tasks = []
        num = 100
        url = f"{self._gis._portal.resturl}portals/self/allScheduledTasks"
        params = {"f": "json", "start": 1, "num": num}
        if item:
            params["itemId"] = item.itemid
        if not active is None:
            params["active"] = active
        if user:
            params["userFilter"] = user.username
        if types:
            params["types"] = types
        res = self._con.get(url, params)
        start = res["nextStart"]
        _tasks.extend(res["tasks"])
        while start != -1:
            params["start"] = start
            params["num"] = num
            res = self._con.get(url, params)
            if len(res["tasks"]) == 0:
                break
            _tasks.extend(res["tasks"])
            start = res["nextStart"]
        return _tasks

    # ----------------------------------------------------------------------
    @property
    def _location_tracking(self):
        """
        The manager for Location Sharing. See :class:`~arcgis.apps.tracker.LocationTrackingManager'.
        """
        return LocationTrackingManager(self._gis)

    # ----------------------------------------------------------------------
    @property
    def social_providers(self):
        """
        Accesses the :class:`~arcgis.gis.admin.SocialProviders` resource to
        allow for the setting and configuration of the social providers
        for the organization.
        """
        if self._sp is None:
            from arcgis.gis.admin._socialproviders import SocialProviders

            self._sp = SocialProviders(gis=self._gis)
        return self._sp

    # ----------------------------------------------------------------------
    @property
    def metadata(self):
        """
        Accesses the :class:`~arcgis.gis.admin.MetadataManager` which
        provides a set of tools to work with the organization's metadata
        settings.
        """
        if self._metadata is None:
            from arcgis.gis.admin._metadata import MetadataManager

            self._metadata = MetadataManager(gis=self._gis)
        return self._metadata

    # ----------------------------------------------------------------------
    @property
    def organizations(self):
        """Provides access to the :class:`~arcgis.gis.kubernetes.KubeOrganizations`
        object to work with the organization's settings."""
        if self._organizations is None:
            from ._organizations import KubeOrganizations

            url = f"{self._url}/orgs"
            self._organizations = KubeOrganizations(url=url, gis=self._gis)
        return self._organizations

    # ----------------------------------------------------------------------
    @property
    def services(self):
        """Provides access to the Kubernetes
        :class:`~arcgis.gis.kubernetes.ServicesManager` object for the site
        """
        if self._services is None:
            from ._services import ServicesManager

            url = f"{self._url}/services"
            self._services = ServicesManager(url, gis=self._gis)
        return self._services

    # ----------------------------------------------------------------------
    @property
    def services_catalog(self):
        """Provides access to the kubernetes
        :class:`~arcgis.gis.kubernetes._server.KubeServiceDirectory` work with the
        services on the site.
        """
        if self._catalog is None:
            from arcgis.gis.kubernetes._server import KubeServiceDirectory

            url = f"{self._url.replace('/admin', '/rest')}/services"
            self._catalog = KubeServiceDirectory(url, gis=self._gis)
        return self._catalog

    # ----------------------------------------------------------------------
    @property
    def uploads(self):
        """Gets the :class:`~arcgis.gis.kubernetes.Uploads` object to work
        with the site uploads."""
        if self._uploads is None:
            from ._uploads import Uploads

            url = self._url + "/uploads"
            self._uploads = Uploads(url=url, gis=self._con, initialize=True)
        return self._uploads

    # ----------------------------------------------------------------------
    @property
    def security(self) -> "KubeSecurity":
        """
        Gets a :class:`~arcgis.gis.kubernetes.KubeSecurity` object to work
        with the site's security settings

        :return:
            :class:`~arcgis.gis.kubernetes.KubeSecurity` object.
        """
        if self._security is None:
            from arcgis.gis.kubernetes._admin._security import KubeSecurity

            url = self._url + "/security"
            self._security = KubeSecurity(url=url, gis=self._gis)
        return self._security

    # ----------------------------------------------------------------------
    @property
    def collaborations(self):
        """
        The collaborations resource lists all collaborations in which a
        portal participates
        """
        if self._collaborations is None:
            from arcgis.gis.admin._collaboration import CollaborationManager

            self._collaborations = CollaborationManager(gis=self._gis)
        return self._collaborations

    # ----------------------------------------------------------------------
    @property
    def webhooks(self):
        """Provides access to the organiztion's
        :class:`Webhook Manager <arcgis.gis.admin.WebHookManager>`."""
        if self._whm is None and self._gis.version >= [6, 4]:
            from arcgis.gis.admin._wh import WebhookManager

            url = self._gis._portal.resturl + "portals/self/webhooks"
            self._whm = WebhookManager(url=url, gis=self._gis)
        return self._whm



import numpy as np
import warnings


def extract_zipfile(filepath, filename, remove=False):
    """Function to extract the contents of a zip file
    Args:
        filepath: absolute path to the file directory.
        filename: name of the zip file to be extracted.
        remove: default=False, removes the original zip file
                after extracting the contents if True
    """
    import os, zipfile

    with zipfile.ZipFile(os.path.join(filepath, filename), "r") as zip_ref:
        zip_ref.extractall(filepath)
    if remove:
        os.remove(os.path.join(filepath, filename))


def arcpy_localization_helper(msg, id, msg_type="ERROR", param=None):
    try:
        import arcpy

        arcpy.AddIDMessage(msg_type, id, param)
        if msg_type == "ERROR":
            try:
                exit()
            except:
                pass
    except:
        return msg


def chips_to_batch(chips, model_height, model_width, batch_size=1):
    dtype = np.float32
    band_count = 3
    if len(chips) != 0:
        dtype = chips[0].dtype

    batch = np.zeros(
        shape=(batch_size, band_count, model_height, model_width),
        dtype=dtype,
    )
    for b in range(batch_size):
        if b < len(chips):
            batch[b, :, :model_height, :model_height] = chips[b]

    return batch


def check_imbalance(total_sample, unique_sample, class_imbalance_pct, stratify):
    """Function to check class imabalance in the data
    Args:
        total_sample: Total number of samples combining all the class(es).
        unique_sample: Class(es) in the data.
        class_imbalance_pct: Percentage of data for each class to consider for imbalance
    """
    imabalanced_class_list = []
    for sample in unique_sample:
        if (total_sample == sample).sum() < len(total_sample) * class_imbalance_pct:
            imabalanced_class_list.append(sample)

    if stratify == True and len(imabalanced_class_list) > 0:
        warnings.warn(
            f'We see a class imbalance in the dataset. The class(es) {",".join(imabalanced_class_list)} does not have enough data points in your dataset.'
        )
    elif stratify == False and len(imabalanced_class_list) > 0:
        warnings.warn(
            f'We see a class imbalance in the dataset. The class(es) {",".join(imabalanced_class_list)} does not have enough data points in your dataset. Although, class imbalance cannot be overcome easily, adding the parameter stratify = True will to a certain extent help get over this problem.'
        )



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# sabl_cascade_rcnn_r101_fpn_1x_coco.py, box AP = 43.0
_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"
# model settings
model = dict(
    pretrained="torchvision://resnet101",
    backbone=dict(depth=101),
    roi_head=dict(
        bbox_head=[
            dict(
                type="SABLHead",
                num_classes=80,
                cls_in_channels=256,
                reg_in_channels=256,
                roi_feat_size=7,
                reg_feat_up_ratio=2,
                reg_pre_kernel=3,
                reg_post_kernel=3,
                reg_pre_num=2,
                reg_post_num=1,
                cls_out_channels=1024,
                reg_offset_out_channels=256,
                reg_cls_out_channels=256,
                num_cls_fcs=1,
                num_reg_fcs=0,
                reg_class_agnostic=True,
                norm_cfg=None,
                bbox_coder=dict(
                    type="BucketingBBoxCoder", num_buckets=14, scale_factor=1.7
                ),
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0
                ),
                loss_bbox_reg=dict(type="SmoothL1Loss", beta=0.1, loss_weight=1.0),
            ),
            dict(
                type="SABLHead",
                num_classes=80,
                cls_in_channels=256,
                reg_in_channels=256,
                roi_feat_size=7,
                reg_feat_up_ratio=2,
                reg_pre_kernel=3,
                reg_post_kernel=3,
                reg_pre_num=2,
                reg_post_num=1,
                cls_out_channels=1024,
                reg_offset_out_channels=256,
                reg_cls_out_channels=256,
                num_cls_fcs=1,
                num_reg_fcs=0,
                reg_class_agnostic=True,
                norm_cfg=None,
                bbox_coder=dict(
                    type="BucketingBBoxCoder", num_buckets=14, scale_factor=1.5
                ),
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0
                ),
                loss_bbox_reg=dict(type="SmoothL1Loss", beta=0.1, loss_weight=1.0),
            ),
            dict(
                type="SABLHead",
                num_classes=80,
                cls_in_channels=256,
                reg_in_channels=256,
                roi_feat_size=7,
                reg_feat_up_ratio=2,
                reg_pre_kernel=3,
                reg_post_kernel=3,
                reg_pre_num=2,
                reg_post_num=1,
                cls_out_channels=1024,
                reg_offset_out_channels=256,
                reg_cls_out_channels=256,
                num_cls_fcs=1,
                num_reg_fcs=0,
                reg_class_agnostic=True,
                norm_cfg=None,
                bbox_coder=dict(
                    type="BucketingBBoxCoder", num_buckets=14, scale_factor=1.3
                ),
                loss_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0
                ),
                loss_bbox_cls=dict(
                    type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0
                ),
                loss_bbox_reg=dict(type="SmoothL1Loss", beta=0.1, loss_weight=1.0),
            ),
        ]
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/sabl/sabl_cascade_rcnn_r101_fpn_1x_coco/sabl_cascade_rcnn_r101_fpn_1x_coco-2b83e87c.pth"



""" Defines the Project object.
"""
import arcgis
from arcgis.features import FeatureLayer, Table
from arcgis.gis import Group
from arcgis._impl.common._utils import _lazy_property
from warnings import warn
import json

from ._schemas import *
from .managers import *
from arcgis.apps.workforce.exceptions import WorkforceError


class Project:
    """
    A Workforce Project

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    item                   Required :class:`~arcgis.gis.Item`. The item that
                           the contains the project.

                           For a version 1 Workforce project, this is an item of type
                           `Workforce Project`. For a version 2 Workforce project, this is an
                           item of type `Feature Service` with typeKeyword `Workforce Project`
    ==================     ====================================================================

    .. code-block:: python

        # Get a Project and search the assignments and workers.

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        assignments = project.assignments.search()
        workers = project.workers.search()

        # Create v1 "Classic" Workforce project and v2 "offline-enabled" project
        v1_project = arcgis.apps.workforce.create_project('v1_project', major_version=1)
        v2_project = arcgis.apps.workforce.create_project('v2_project', major_version=2)


    """

    def __init__(self, item):
        """
        :param item: The project's arcigs.gis.Item
        """
        self.gis = item._gis
        if "Workforce Project" in item.typeKeywords:
            self._item = item
        else:
            raise WorkforceError("Incorrect item passed into Project class")
        if self._is_v2_project:
            self._item_data = item.properties
        else:
            self._item_data = item.get_data()
        self._assignment_schema = AssignmentSchema(self.assignments_layer)
        if self._supports_tracks:
            self._track_schema = TrackSchema(self.tracks_layer)
        else:
            self._track_schema = None
        self._worker_schema = WorkerSchema(self.workers_layer)
        if self._is_v2_project:
            self._assignment_types = AssignmentTypeSchema(self.assignment_types_table)
            self._integration_schema = IntegrationSchema(self.integrations_table)
        self._dispatcher_schema = DispatcherSchema(self.dispatchers_layer)
        self._update_cached_objects()

    def _update_cached_assignment_types(self):
        """
        Updates the cached assignment types
        """
        self._cached_assignment_types = {
            a.code: a for a in self.assignment_types.search()
        }

    def _update_cached_objects(self):
        """
        Caches the types, workers, and dispatchers for quicker assignment creation when querying
        Should be called when querying assignments
        """
        self._update_cached_assignment_types()
        self._cached_workers = {w.id: w for w in self.workers.search()}
        self._cached_dispatchers = {d.id: d for d in self.dispatchers.search()}
        for d in self._cached_dispatchers.values():
            if d.user_id == self.gis.users.me.username:
                self._cached_dispatcher = d
                break
        else:
            raise WorkforceError(
                "'{}' is not a dispatcher, please authenticate as a dispatcher".format(
                    self.gis.users.me.username
                )
            )

    def __str__(self):
        return self.title

    def __repr__(self):
        return "<Project {}>".format(self.title)

    @staticmethod
    def _delete_item_with_related_views(item):
        """
        Recursively fetches all related feature service items and then will delete any child services before deleting the parent service(s).
        :param item: An item
        """
        try:
            # Related Items API might not exist for older Enterprise versions
            related_items = item.related_items("Service2Service", "forward")
        except Exception:
            related_items = []
        for v in related_items:
            Project._delete_item_with_related_views(v)
        item.protect(False)
        item.delete()

    def delete(self):
        """
        Deletes the project, group, folder, layers, and webmaps.
        Assumes the currently signed in user owns the project or is an admin.
        """
        title = self.title
        owner = self._item.owner
        self._delete_item_with_related_views(self.assignments_item)
        if self._supports_tracks:
            self.tracks_item.protect(False)
            self.tracks_item.delete()
        if not self._is_v2_project:
            self.workers_item.protect(False)
            self.workers_item.delete()
            self.dispatchers_item.protect(False)
            self.dispatchers_item.delete()
            self._item.protect(False)
            self._item.delete()
        self.dispatcher_webmap.item.protect(False)
        self.dispatcher_webmap.item.delete()
        self.worker_webmap.item.protect(False)
        self.worker_webmap.item.delete()
        self.group.protected = False
        self.group.delete()
        for folder in self.gis.users.get(owner).folders:
            if folder["title"] == title:
                self.gis.content.delete_folder(folder["title"], owner=owner)

    def _update_data(self):
        # this function is used by v1 projects only
        self._item.update({"text": json.dumps(self._item_data)})

    def update(self, summary=None):
        """
        Updates the project on the server

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        summary                  Optional :class:`String`. The summary of the project.
        ==================     ====================================================================
        """
        item_properties = {}
        if summary:
            item_properties["snippet"] = summary
        self._item.update(item_properties)
        if self._is_v2_project:
            self.gis.content.get(self.dispatcher_web_map_id).update(item_properties)
            self.gis.content.get(self.worker_web_map_id).update(item_properties)

    @property
    def _supports_tracks(self):
        return not self._is_v2_project and "tracks" in self._item_data

    @property
    def _is_v2_project(self):
        try:
            return (
                self._item.type == "Feature Service"
                and int(self._item.properties["workforceProjectVersion"].split(".")[0])
                >= 2
            )
        except:
            return False

    @property
    def _tracking_enabled(self):
        if self._supports_tracks:
            return self._item_data["tracks"]["enabled"]

    @_tracking_enabled.setter
    def _tracking_enabled(self, value):
        if self._supports_tracks:
            self._item_data["tracks"]["enabled"] = value
            self._update_data()

    @property
    def _tracking_interval(self):
        if self._supports_tracks:
            return self._item_data["tracks"]["updateInterval"]

    @_tracking_interval.setter
    def _tracking_interval(self, value):
        if self._supports_tracks:
            self._item_data["tracks"]["updateInterval"] = value
            self._update_data()

    @property
    def integrations(self):
        """The :class:`~arcgis.apps.workforce.managers.AssignmentIntegrationManager` for the project"""
        return AssignmentIntegrationManager(self)

    @property
    def id(self):
        """The item id of the project"""
        return self._item["id"]

    @property
    def title(self):
        """Gets the title of the project"""
        return self._item["title"]

    @property
    def summary(self):
        """The title of the project"""
        return self._item["snippet"]

    @summary.setter
    def summary(self, value):
        self._item["snippet"] = value

    @property
    def owner_user_id(self):
        """The user id of the project owner."""
        return self._item["owner"]

    @property
    def version(self):
        """The version of the project"""
        if self._is_v2_project:
            return self._item_data["workforceProjectVersion"]
        else:
            return self._item_data["version"]

    @_lazy_property
    def assignments_item(self):
        """The assignments :class:`~arcgis.gis.Item`"""
        if self._is_v2_project:
            # this is the same item as workers_item, dispatchers_item for a v2 project - each points to the one FS
            return self.gis.content.get(self._item.id)
        else:
            return self.gis.content.get(self._item_data["assignments"]["serviceItemId"])

    @property
    def assignments_layer_url(self):
        """The assignments feature layer url"""
        if self._is_v2_project:
            return self._item.url + "/0"
        else:
            return self._item_data["assignments"]["url"]

    @_lazy_property
    def assignment_types_item(self):
        """The assignment types :class:`~arcgis.gis.Item`"""
        if self._is_v2_project:
            return self.gis.content.get(self._item.id)
        else:
            warn(
                "This Workforce Project does not have an assignment types item",
                WorkforceWarning,
            )

    @property
    def assignment_types_table_url(self):
        """The assignment types table url"""
        if self._is_v2_project:
            return self._item.url + "/3"
        else:
            warn(
                "This Workforce Project does not have an assignment types table",
                WorkforceWarning,
            )

    @_lazy_property
    def dispatchers_item(self):
        """The dispatchers :class:`~arcgis.gis.Item`"""
        if self._is_v2_project:
            return self.gis.content.get(self._item.id)
        else:
            return self.gis.content.get(self._item_data["dispatchers"]["serviceItemId"])

    @property
    def dispatchers_layer_url(self):
        """The dispatchers layer url"""
        if self._is_v2_project:
            return self._item.url + "/2"
        else:
            return self._item_data["dispatchers"]["url"]

    @_lazy_property
    def integrations_table_url(self):
        """The integrations table url :class:`~arcgis.features.Table`"""
        if self._is_v2_project:
            return self._item.url + "/4"
        else:
            warn(
                "This Workforce Project does not have an integrations table",
                WorkforceWarning,
            )

    @_lazy_property
    def tracks_item(self):
        """The tracks :class:`~arcgis.gis.Item`"""
        if self._supports_tracks:
            return self.gis.content.get(self._item_data["tracks"]["serviceItemId"])
        else:
            warn("This Workforce Project does not support tracks.", WorkforceWarning)

    @property
    def tracks_layer_url(self):
        """The tracks feature layer url"""
        if self._supports_tracks:
            return self._item_data["tracks"]["url"]
        else:
            warn("This Workforce Project does not support tracks.", WorkforceWarning)

    @_lazy_property
    def workers_item(self):
        """The workers :class:`~arcgis.gis.Item`"""
        if self._is_v2_project:
            return self.gis.content.get(self._item.id)
        else:
            return self.gis.content.get(self._item_data["workers"]["serviceItemId"])

    @property
    def workers_layer_url(self):
        """The workers feature layer url"""
        if self._is_v2_project:
            return self._item.url + "/1"
        else:
            return self._item_data["workers"]["url"]

    @property
    def dispatcher_web_map_id(self):
        """The dispatcher webmap item id"""
        if self._is_v2_project:
            # not all systems will support this new "Workforce2MapFeatureService" so we try/except. If the rel does
            # not exist, we can get the webmap out of the metadata
            try:
                related_items = self._item.related_items(
                    "WorkforceMap2FeatureService", "reverse"
                )
                for item in related_items:
                    if "Workforce Dispatcher" in item.typeKeywords:
                        return item.id
                return self._item_data["workforceDispatcherMapId"]
            except Exception:
                return self._item_data["workforceDispatcherMapId"]
        else:
            return self._item_data["dispatcherWebMapId"]

    @property
    def worker_web_map_id(self):
        """The worker webmap item id"""
        if self._is_v2_project:
            try:
                related_items = self._item.related_items(
                    "WorkforceMap2FeatureService", "reverse"
                )
                for item in related_items:
                    if "Workforce Worker" in item.typeKeywords:
                        return item.id
                return self._item_data["workforceWorkerMapId"]
            except Exception:
                return self._item_data["workforceWorkerMapId"]
        else:
            return self._item_data["workerWebMapId"]

    @property
    def group_id(self):
        """The group id that all project items are part of"""
        if self._is_v2_project:
            return self._item_data["workforceProjectGroupId"]
        else:
            return self._item_data["groupId"]

    @_lazy_property
    def owner(self):
        """The owner :class:`~arcgis.gis.User` of the project"""
        return self.gis.users.get(self.owner_user_id)

    @_lazy_property
    def assignments_layer(self):
        """The assignments :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.assignments_layer_url, self.gis)

    @_lazy_property
    def dispatchers_layer(self):
        """The dispatchers :class:`~arcgis.features.FeatureLayer`"""
        if self._is_v2_project:
            return Table(self.dispatchers_layer_url, self.gis)
        else:
            return FeatureLayer(self.dispatchers_layer_url, self.gis)

    @_lazy_property
    def assignment_types_table(self):
        """The assignment types :class:`~arcgis.features.Table`"""
        if self._is_v2_project:
            return Table(self.assignment_types_table_url, self.gis)
        else:
            warn(
                "This Workforce Project does not have an assignment types table",
                WorkforceWarning,
            )

    @_lazy_property
    def integrations_table(self):
        """The integrations :class:`~arcgis.features.Table`"""
        if self._is_v2_project:
            return Table(self.integrations_table_url, self.gis)
        else:
            warn(
                "This Workforce Project does not have an integrations table",
                WorkforceWarning,
            )

    @_lazy_property
    def tracks_layer(self):
        """The tracks :class:`~arcgis.features.FeatureLayer`"""
        if self._supports_tracks:
            return FeatureLayer(self.tracks_layer_url, self.gis)
        else:
            warn("This Workforce Project does not support tracks.", WorkforceWarning)

    @_lazy_property
    def workers_layer(self):
        """The workers :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.workers_layer_url, self.gis)

    @_lazy_property
    def dispatcher_webmap(self):
        """The dispatcher :class:`~arcgis.mapping.WebMap` for the project"""
        return arcgis.mapping.WebMap(self.gis.content.get(self.dispatcher_web_map_id))

    @_lazy_property
    def worker_webmap(self):
        """The worker :class:`~arcgis.mapping.WebMap` for the project"""
        return arcgis.mapping.WebMap(self.gis.content.get(self.worker_web_map_id))

    @_lazy_property
    def group(self):
        """The :class:`~arcgis.gis.Group` that the project resources are part of"""
        return Group(self.gis, self.group_id)

    @property
    def assignments(self):
        """The :class:`~arcgis.apps.workforce.managers.AssignmentManager` for the project"""
        return AssignmentManager(self)

    @property
    def workers(self):
        """The :class:`~arcgis.apps.workforce.managers.WorkerManager` for the project"""
        return WorkerManager(self)

    @property
    def dispatchers(self):
        """The :class:`~arcgis.apps.workforce.managers.DispatcherManager` for the project"""
        return DispatcherManager(self)

    @property
    def tracks(self):
        """The :class:`~arcgis.apps.workforce.managers.TrackManager` for the project"""
        if self._supports_tracks:
            return TrackManager(self)
        else:
            warn("This Workforce Project does not support tracks.", WorkforceWarning)

    @property
    def assignment_types(self):
        """The :class:`~arcgis.apps.workforce.managers.AssignmentTypeManager` for the project"""
        return AssignmentTypeManager(self)



from __future__ import annotations
from typing import Any
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS


class Job(_BaseKube):
    """A Kubernetes asynchronous opertion"""

    _properties = None

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict[str, Any]:
        """
        returns the object properties
        """
        self._properties = None
        if self._properties is None:
            self._init()
        return self._properties


###########################################################################
class JobManager(_BaseKube):
    """
    Provides access to the jobs resources defined on the ArcGIS
    Enterprise.
    """

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ---------------------------------------------------------------------
    def job(self, job_id):
        """
        This resource returns the progress and status messages of an asynchronous
        job. Updated progress information can be acquired by periodically querying this operation.
        """
        url = f"{self._url}/{job_id}"
        params = {"f": "json"}
        return self._con.get(url, params)



from .._utils.env import _LAMBDA_TEXT_CLASSIFICATION

if not _LAMBDA_TEXT_CLASSIFICATION:
    from ._ssd import SingleShotDetector
    from ._unet import UnetClassifier
    from ._retinanet import RetinaNet
    from ._classifier import FeatureClassifier
    from ._pspnet import PSPNetClassifier
    from ._maskrcnn import MaskRCNN
    from ._deeplab import DeepLab
    from ._pointcnnseg import PointCNN
    from ._yolov3 import YOLOv3
    from ._layer_learner import FullyConnectedNetwork
    from ._machine_learning import MLModel
    from ._model_extension import ModelExtension
    from ._faster_rcnn import FasterRCNN
    from ._superres import SuperResolution
    from ._hed import HEDEdgeDetector
    from ._bdcn import BDCNEdgeDetector
    from ._image_captioner import ImageCaptioner
    from ._cyclegan import CycleGAN
    from ._tsmodel import TimeSeriesModel
    from ._multi_task_road_extractor import MultiTaskRoadExtractor
    from ._change_detector import ChangeDetector
    from ._pix2pix import Pix2Pix
    from ._mmdetection import MMDetection
    from ._embeddings import Embeddings
    from ._connect_net import ConnectNet
    from ._mmsegmentation import MMSegmentation
    from ._siammask import SiamMask, Track
    from ._deepsort import DeepSort
    from ._auto_ml import AutoML
    from ._pix2pix_hd import Pix2PixHD
    from ._auto_dl import AutoDL, ImageryModel
    from ._max_deeplab import MaXDeepLab
    from ._wnet_cgan import WNet_cGAN
    from ._detreg_detector import DETReg
    from ._RandLANet import RandLANet
    from ._efficientdet import EfficientDet
    from ._SQNSeg import SQNSeg
    from ._psetae import PSETAE
    from ._mmdet3d import MMDetection3D



try:
    from .._data import _raise_conda_import_error
    import traceback
    from . import _track_processor

    HAS_CONDA = True
except ImportError:
    import_exception = traceback.format_exc()
    HAS_CONDA = False


def get_default_tracker_options():
    tracker_options = {
        "enable_post_processing": True,
        "detection_interval": 1,
        "detection_threshold": 0.0,
        "detect_track_failure": False,
        "recover_track": False,
        "stab_period": 6,
        "detect_fail_interval": 0,
        "min_obj_size": 10,
        "template_history": 25,
        "status_history": 60,
        "status_fail_threshold": 0.6,
        "search_period": 60,
        "knn_distance_ratio": 0.75,
        "recover_conf_threshold": 0.1,
        "recover_iou_threshold": 0.1,
    }

    return tracker_options


class TrackProcessor:

    """
    Creates TrackProcessor Object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    processor_options       Optional dictionary. A dictionary with
                            keys as parameter names and values as
                            parameter values.

                            "detection_interval" refers to
                            the interval in frames at which the detector
                            is invoked.

                            "detection_threshold" refers to
                            the lower threshold for selecting the
                            detections.

                            "detect_track_failure" refers to
                            the flag which enables/disables the logic
                            to detect whether the object appearance
                            has changed detection.

                            "recover_track" refers to the flag which
                            enables/disables track recovery post failure.

                            "stab_period" refers to the number of frames
                            after which post processing starts.

                            "detect_fail_interval" refers to the number
                            of frames after which to detect track failure.

                            "min_obj_size" refers to the size in pixels
                            below which tracking is assumed to have
                            failed.

                            "template_history" refers to the number of
                            frames before the current frame at which
                            template image is fetched.

                            "status_history" refers to the
                            number of frames over which status of the
                            track is used to detect track failure.

                            "status_fail_threshold" refers to the
                            threshold for the ratio between number
                            of frames for which object is searched
                            for and the total number of frames which
                            needs to be crossed for track failure
                            detection.

                            "search_period" refers to the
                            number of frames for which object is
                            searched for before declaring object is
                            lost.

                            "knn_distance_ratio" refers to the
                            threshold for ratio of the distances between
                            template descriptor and the two best matched
                            detection descriptor, used for filtering
                            best matches.

                            "recover_conf_threshold" refers
                            to the minimum confidence value over which
                            recovery logic is enabled.

                            "recover_iou_threshold" refers to the minimum
                            overlap between template and detection for
                            successful recovery.
    =====================   ===========================================

    :return: `TrackProcessor` Object
    """

    def __init__(
        self,
        processor_options={
            "detect_track_failure": True,
            "recover_track": True,
            "stab_period": 6,
            "detect_fail_interval": 5,
            "min_obj_size": 10,
            "template_history": 25,
            "status_history": 60,
            "status_fail_threshold": 0.6,
            "search_period": 60,
            "knn_distance_ratio": 0.75,
            "recover_conf_threshold": 0.1,
            "recover_iou_threshold": 0.1,
        },
    ):
        if not HAS_CONDA:
            _raise_conda_import_error(import_exception=import_exception)

        self.tracks = []
        params = []
        params.append(processor_options.get("detect_track_failure", True))
        params.append(processor_options.get("recover_track", True))
        params.append(processor_options.get("stab_period", 6))
        params.append(processor_options.get("detect_fail_interval", 5))
        params.append(processor_options.get("min_obj_size", 10))
        params.append(processor_options.get("template_history", 25))
        params.append(processor_options.get("status_history", 60))
        params.append(processor_options.get("status_fail_threshold", 0.6))
        params.append(processor_options.get("search_period", 60))
        params.append(processor_options.get("knn_distance_ratio", 0.75))
        params.append(processor_options.get("recover_conf_threshold", 0.1))
        params.append(processor_options.get("recover_iou_threshold", 0.1))
        _track_processor.create(params)

    def init(self, frame, dets, reset=True):
        """
        Initializes tracks based on the detections returned by detector/
        manually fed to the function.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is used to initialize
                                the objects to track.
        ---------------------   -------------------------------------------
        dets                    Required list. A 1D list with bounding boxes
                                to intialize the tracks.
        ---------------------   -------------------------------------------
        reset                   Optional flag. Indicates whether to reset
                                the tracker and remove all existing tracks
                                before initialization.
        =====================   ===========================================

        :return: None
        """

        # reset = True

        if reset:
            _track_processor.init(frame, dets)
        else:
            _track_processor.add(frame, dets)
        # TODO: remove return - no need
        return None

    def update(self, frame, dets):
        """
        Post-processes the tracks, detects track failure and recovers the
        lost objects if possible.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is the current
                                frame to be used to track the objects.
        ---------------------   -------------------------------------------
        dets                    Required list. A 1D list with state of
                                tracks needed for post processing.
        =====================   ===========================================

        :return: 1D list with updated state of track objects
        """

        dets = _track_processor.update(frame, dets)
        return dets

    def remove(self, tracks_ids):
        """
        Removes the tracks corresponding to track_ids parameter

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        tracks_ids              Required list. List of track ids to be
                                removed.
        =====================   ===========================================

        """
        _track_processor.purge(tracks_ids)
        return None



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# detectors_cascade_rcnn_r50_1x_coco.py, box AP=47.4

_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"

model = dict(
    backbone=dict(
        type="DetectoRS_ResNet",
        conv_cfg=dict(type="ConvAWS"),
        sac=dict(type="SAC", use_deform=True),
        stage_with_sac=(False, True, True, True),
        output_img=True,
    ),
    neck=dict(
        type="RFP",
        rfp_steps=2,
        aspp_out_channels=64,
        aspp_dilations=(1, 3, 6, 1),
        rfp_backbone=dict(
            rfp_inplanes=256,
            type="DetectoRS_ResNet",
            depth=50,
            in_channels=3,
            num_stages=4,
            out_indices=(0, 1, 2, 3),
            frozen_stages=1,
            norm_cfg=dict(type="BN", requires_grad=True),
            norm_eval=True,
            conv_cfg=dict(type="ConvAWS"),
            sac=dict(type="SAC", use_deform=True),
            stage_with_sac=(False, True, True, True),
            pretrained="torchvision://resnet50",
            style="pytorch",
        ),
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/detectors/detectors_cascade_rcnn_r50_1x_coco/detectors_cascade_rcnn_r50_1x_coco-32a10ba0.pth"



import logging
from arcgis.gis import GIS

from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis.server._service import Service
from typing import Optional

_log = logging.getLogger()


class KubeServiceDirectory(_BaseKube):
    """
    A representation of the Kubernetes Hosting Service Directory.

    This is a private method and should not be created by a user.
    """

    _con = None
    _gis = None
    _url = None
    _folders = None
    _folder = None
    _services = None

    def __init__(self, url: str, gis: GIS) -> None:
        """initializer"""
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def report(self, as_html: bool = True, folder: Optional[str] = None):
        """
        Generates a table of Services in the given folder, as a Pandas dataframe.


        """
        import pandas as pd

        pd.set_option("display.max_colwidth", None)
        data = []
        a_template = """<a href="%s?token=%s">URL Link</a>"""
        columns = ["Service Name", "Service URL"]
        if folder is None:
            url = self._url
            res = self._con.get(url, {"f": "json"})
        elif folder.lower() in [f.lower() for f in self.folders]:
            url = "%s/%s" % (self._url, folder)
            res = self._con.get(url, {"f": "json"})
        if "services" in res:
            for s in res["services"]:
                # if s['name'].split('/')[-1].lower() == name.lower():
                url = "%s/%s/%s" % (url, s["name"], s["type"])
                data.append(
                    [
                        s["name"].split("/")[-1],
                        """<a href="%s">Service</a>""" % url,
                    ]
                )

        df = pd.DataFrame(data=data, columns=columns)
        if as_html:
            table = (
                """<div class="9item_container" style="height: auto; overflow: hidden; """
                + """border: 1px solid #cfcfcf; border-radius: 2px; background: #f6fafa; """
                + """line-height: 1.21429em; padding: 10px;">%s</div>"""
                % df.to_html(escape=False, index=False)
            )
            return table.replace("\n", "")
        else:
            return df

    # ----------------------------------------------------------------------
    def get(self, name: str, folder: Optional[str] = None):
        """returns a single service in a folder"""
        if folder is None:
            url = self._url
            res = self._con.get(self._url, {"f": "json"})
        elif folder.lower() in [f.lower() for f in self.folders]:
            url = "%s/%s" % (self._url, folder)
            res = self._con.get("%s/%s" % (self._url, folder), {"f": "json"})
        if "services" in res:
            for s in res["services"]:
                if s["name"].split("/")[-1].lower() == name.lower():
                    return Service(
                        url="%s/%s/%s" % (url, s["name"], s["type"]),
                        server=self._con,
                    )
                del s
        return None

    # ----------------------------------------------------------------------
    def list(self, folder: Optional[str] = None):
        """
        returns a list of services at the given folder
        """
        services = []
        if folder:
            url = "%s/%s" % (self._url, folder)
        else:
            url = self._url
        if folder is None:
            res = self._con.get(url, {"f": "json"})
        elif folder.lower() in [f.lower() for f in self.folders]:
            res = self._con.get(url, {"f": "json"})
        if "services" in res:
            for s in res["services"]:
                try:
                    services.append(
                        Service(
                            url="%s/%s/%s" % (url, s["name"], s["type"]),
                            server=self._con,
                        )
                    )

                except:
                    url = "%s/%s/%s" % (url, s["name"], s["type"])
                    _log.warning("Could not load service: %s" % url)
        return services

    # ----------------------------------------------------------------------
    def find(self, service_name: str, folder: Optional[str] = None):
        """
        finds a service based on it's name in a given folder
        """
        return self.get(name=service_name, folder=folder)

    # ----------------------------------------------------------------------
    @property
    def folders(self):
        """
        returns a list of server folders
        """
        self._properties = None
        if self._is_agol:
            return ["/"]
        else:
            return self.properties["folders"]
        return []

    # ----------------------------------------------------------------------
    def publish_sd(
        self,
        sd_file: str,
        folder: Optional[str] = None,
        service_config: Optional[dict] = None,
    ):
        """
        Publishes a service definition file to ArcGIS Server.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        sd_file                Required string. The service definition file to be uploaded and published.
        ------------------     --------------------------------------------------------------------
        folder                 Optional string. The folder in which to publish the service definition
                               file to.  If this folder is not present, it will be created.  The
                               default is None in which case the service definition will be published
                               to the System folder.
        ------------------     --------------------------------------------------------------------
        service_config         Optional Dict[str, Any]. A set of configuration overwrites that overrides the service definitions defaults.
        ==================     ====================================================================

        :return:
           A boolean indicating success (True) or failure (False).
        """
        import json

        sm = self._gis.admin.services
        catalog = self._gis.admin.services_catalog
        uploads = self._gis.admin.uploads
        if sd_file.lower().endswith(".sd") == False:
            return False
        # catalog = self.content
        if "System" not in catalog.folders:
            return False
        if folder and folder.lower() not in [f.lower() for f in catalog.folders]:
            sm.create_folder(folder)
        service = catalog.get(name="PublishingTools", folder="System")
        if service is None:
            service = catalog.get(name="PublishingToolsEx", folder="System")
        if service is None:
            return False
        status, res = uploads.upload(path=sd_file, description="sd file")
        if status:
            uid = res["item"]["itemID"]
            config = uploads._service_configuration(uid).get("service", {})
            if service_config or folder:
                if service_config:
                    config.update(service_config)
                if folder:
                    if "folderName" in config:
                        config["folderName"] = folder
                res = service.publish_service_definition(
                    in_sdp_id=uid, in_config_overwrite=json.dumps(config)
                )
            else:
                uid = res["item"]["itemID"]
                res = service.publish_service_definition(in_sdp_id=uid)
            return True
        return False



from typing import Collection
import fastai


precondition = True

if fastai.__version__ == "1.0.60":
    import fastai.vision.image
    from fastai.vision.image import Image, plt, image2np
    from .common import ArcGISMSImage

    def show_image(
        img: Image,
        ax: plt.Axes = None,
        figsize: tuple = (3, 3),
        hide_axis: bool = True,
        cmap: str = "binary",
        alpha: float = None,
        **kwargs
    ) -> plt.Axes:
        "Display `Image` in notebook."
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        xtr = dict(cmap=cmap, alpha=alpha, **kwargs)
        if isinstance(img, ArcGISMSImage):
            img.show(ax=ax)
        else:
            ax.imshow(image2np(img.data), **xtr) if (
                hasattr(img, "data")
            ) else ax.imshow(img, **xtr)
        if hide_axis:
            ax.axis("off")
        return ax

    #
    fastai.vision.learner.show_image = show_image
    fastai.vision.image.show_image = show_image

    import fastai.data_block
    from fastai.core import array

    def process(self, ds: Collection):
        ds.items = array([self.process_one(item) for item in ds.items], dtype=object)

    fastai.data_block.PreProcessor.process = process



from pathlib import Path
import json
from ._model_extension import ModelExtension
from ._arcgis_model import _EmptyData

try:
    from fastai.vision import flatten_model
    import torch
    from fastai.torch_core import split_model_idx
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._arcgis_model import _resnet_family, _vgg_family
    from ._timm_utils import filter_timm_models
    from ._hed_utils import DDPCallback

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False


class CustomBDCN:
    """
    Create class with following fixed function names and the number of arguents to train your model from external source
    """

    try:
        import torch
        from torchvision import models
        from arcgis.learn.models import _bdcn_utils as bdcn
    except:
        pass

    def get_model(self, data, backbone=None, **kwargs):
        """
        In this fuction you have to define your model with following two arguments!

        """
        pretrained_backbone = kwargs.get("pretrained_backbone", True)

        if backbone is None:
            self._backbone = self.models.vgg19
        elif type(backbone) is str:
            if hasattr(self.models, backbone):
                self._backbone = getattr(self.models, backbone)
            elif hasattr(self.models.detection, backbone):
                self._backbone = getattr(self.models.detection, backbone)
            elif "timm:" in backbone:
                import timm

                bckbn = backbone.split(":")[1]
                if hasattr(timm.models, bckbn):
                    self._backbone = getattr(timm.models, bckbn)
        else:
            self._backbone = backbone

        model = self.bdcn._BDCNModel(
            self._backbone, data.chip_size, pretrained=pretrained_backbone
        )

        return model

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        return model_input_batch, model_target_batch

    def transform_input(self, xb):
        return xb

    def transform_input_multispectral(self, xb):
        return xb

    def loss(self, model_output, *model_target):
        final_loss = self.bdcn.bdcn_loss(model_output, *model_target)

        return final_loss

    def post_process(self, pred, thres=0.5, thinning=True, prob_raster=False):
        """
        In this function you have to return list with appended output for each image in the batch with shape [C=1,H,W]!

        """

        from skimage.morphology import skeletonize, binary_dilation
        import numpy as np

        post_processed_pred = []
        if isinstance(pred, list):
            pred = pred[-1]
        if prob_raster:
            return pred
        elif thinning:
            for p in pred:
                p = self.torch.unsqueeze(
                    self.torch.tensor(
                        skeletonize(
                            binary_dilation(
                                np.squeeze((p >= thres).byte().cpu().numpy())
                            )
                        )
                    ),
                    dim=0,
                )
                post_processed_pred.append(p)
            return post_processed_pred
        else:
            return (pred >= thres).byte()


class BDCNEdgeDetector(ModelExtension):
    """
    Model architecture from https://arxiv.org/pdf/1902.10903.pdf.
    Creates a :class:`~arcgis.learn.BDCNEdgeDetector` model

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone convolutional neural network
                            model used for feature extraction, which is `vgg19` by
                            default.
                            Supported backbones: ResNet, Vgg family and specified Timm
                            models(experimental support) from :func:`~arcgis.learn.BDCNEdgeDetector.backbones`.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.BDCNEdgeDetector` Object
    """

    def __init__(self, data, backbone="vgg19", pretrained_path=None):
        self._check_dataset_support(data)
        backbone_name = backbone if type(backbone) is str else backbone.__name__
        if backbone_name not in self.supported_backbones:
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, CustomBDCN, backbone, pretrained_path)

        if getattr(self, "_multigpu_training", False):
            self.learn.callbacks.append(DDPCallback(self.learn, self._rank_distributed))
        self._freeze()

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        count = 0
        count_strided_conv = 0
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue

            for p in i.parameters():
                p.requires_grad = False

            if isinstance(i, torch.nn.MaxPool2d):
                count += 1
                if count == 3:
                    break
            if isinstance(i, torch.nn.Conv2d):
                if i.stride[0] == 2:
                    count_strided_conv += 1
                    if count_strided_conv == 4:
                        break

        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy", "f1_score"]

    @property
    def _is_edge_detection(self):
        return True

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return BDCNEdgeDetector._supported_backbones()

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return BDCNEdgeDetector._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(
            [
                "*dpn*",
                "*inception*",
                "*nasnet*",
                "*tf_efficientnet_cc*",
                "*repvgg*",
                "*resnetblur*",
                "*selecsls*",
                "*tresnet*",
                "*hrnet*",
                "*rexnet*",
                "*mixnet*",
                "*ghostnet*",
            ]
        )
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))
        return [*_resnet_family, *_vgg_family] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return BDCNEdgeDetector._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Classified_Tiles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.BDCNEdgeDetector` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.BDCNEdgeDetector` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        backbone = emd["ModelParameters"]["backbone"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=emd["ImageHeight"],
            )
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data.classes = ["background"]
            for k, v in class_mapping.items():
                data.classes.append(v)
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = emd["DatasetType"]

        return cls(data, backbone, pretrained_path=str(model_file))

    def compute_precision_recall(self, thresh=0.5, buffer=3, show_progress=True):
        """
        Computes precision, recall and f1 score on validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability on which
                                the detection will be considered edge pixel.
        ---------------------   -------------------------------------------
        buffer                  Optional int. pixels in neighborhood to
                                consider true detection.
        =====================   ===========================================

        :return: `dict`
        """

    def show_results(self, rows=5, thresh=0.5, thinning=True, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """



import atexit
import json
import arcgis
from arcgis._impl.common._deprecate import deprecated

_DASHBOARD_VERSION = 27

_created_dashboards = []


@deprecated(deprecated_in="2.0.1", removed_in=None, current_version="2.0.1")
class Dashboard(object):
    """
    Creates a :class:`~arcgis.apps.dashboard.Dashboard` Object.

    :return:
        :class:`~arcgis.apps.dashboard.Dashboard` object
    """

    def __init__(self):
        self.elements = []

        self._theme = "light"

        self._header = None
        self._side_panel = None
        self._layout = {}
        self._widgets = []

        # if item is not None:
        #     return _from_dashboard(item)

    def save(
        self, title, description="", summary="", tags=None, gis=None, overwrite=False
    ):
        """
        Publishes a Dashboard Object.

        =========================   ===========================================
        **Parameter**                **Description**
        -------------------------   -------------------------------------------
        title                       Required string. Title or Caption for the
                                    Dashboard.
        -------------------------   -------------------------------------------
        description                 Optional string. Description for the Dashboard.
        -------------------------   -------------------------------------------
        summary                     Optional string. Summary of the Dashboard.
        -------------------------   -------------------------------------------
        tags                        Optional string. Comma separated tags.
        -------------------------   -------------------------------------------
        gis                         Optional :class:`~arcgis.gis.GIS` to publish dashboard.
                                    By default uses active gis.
        -------------------------   -------------------------------------------
        overwrite                   Optional Boolean.
                                    Overwrite existing dashboard.
        =========================   ===========================================
        """
        if not title:
            raise Exception("Please specify a title.")

        self.title = title
        self.description = description

        self.summary = summary
        self.tags = tags if tags else ""
        return self._publish(gis=gis, overwrite=overwrite)

    @property
    def theme(self):
        return self._theme

    @theme.setter
    def theme(self, value):
        self._theme = value
        if value not in ["light", "dark"]:
            self._theme = "light"

    @property
    def header(self):
        """
        :return:
            :class:`~arcgis.apps.dashboard.Header` object
        """
        return self._header

    @header.setter
    def header(self, value):
        """
        Set the header object
        """
        self._header = value

    @property
    def side_panel(self):
        """
        :return:
            :class:`~arcgis.apps.dashboard.SidePanel` object
        """
        return self._side_panel

    @side_panel.setter
    def side_panel(self, value):
        """
        Set the Side Panel object
        """
        self._side_panel = value

    def _convert_to_json(self):
        json_data = {
            "version": _DASHBOARD_VERSION,
            "widgets": [element._convert_to_json() for element in self.elements],
            "settings": {"maxPaginationRecords": 50000, "allowElementResizing": False},
            "mapOverrides": {"trackedFeatureRadius": 60},
            "theme": self.theme,
            "themeOverrides": {},
            "numberPrefixOverrides": [],
            "layout": self._layout,
            "authoringApp": "ArcGIS API for Python",
            "authoringAppVersion": arcgis.__version__,
            "typeKeywords": "Python",
        }

        if self.header:
            json_data["headerPanel"] = self.header._convert_to_json()

        if self.side_panel:
            json_data["leftPanel"] = self.side_panel._convert_to_json()

        # print(json_data)
        return json_data

    @property
    def layout(self):
        """
        :return: Layout of the dashboard
        """
        return self._layout

    @layout.setter
    def layout(self, value):
        """
        Set the layout of the dashboard, using add_row and add_column functions.
        """
        self.elements = value["widgets"]
        del value["widgets"]
        self._layout = {"rootElement": value}

    # @property
    # def widgets(self):
    #     """
    #     :return: widgets of the dashboard
    #     """
    #     return self._widgets

    def _repr_html_(self):
        url = self._dash_publish()
        return f"""<iframe src={url} width=900 height=300>"""

    def _dash_publish(self):
        gis = arcgis.env.active_gis
        import random
        import string

        letters = string.ascii_lowercase
        title = "".join(random.choice(letters) for i in range(10))
        summary = "".join(random.choice(letters) for i in range(10))

        db = Dashboard()

        db.title = title
        db.summary = summary
        db.description = ""
        db.summary = summary
        db.tags = ""

        db._layout = self._layout
        db.elements = self.elements
        db = db._publish(gis)
        _created_dashboards.append((gis, db))
        url = f"{gis.url}/apps/opsdashboard/index.html#/{db.itemid}"

        return url

    def _publish(self, gis=None, overwrite=False):
        if gis is None:
            gis = arcgis.env.active_gis

        items = gis.content.search(f"title:{self.title}", "Dashboard")
        for item in items:
            if item.title.lower() == self.title.lower() and overwrite is False:
                raise Exception(
                    "A dashboard with same name already exists, to continue set `overwrite` = True"
                )
            elif item.title.lower() == self.title.lower():
                item.delete(force=True)

        return gis.content.add(
            {
                "type": "Dashboard",
                "description": self.description,
                "title": self.title,
                "overwrite": str(overwrite).lower(),
                "text": json.dumps(self._convert_to_json()),
            }
        )

    def _from_dashboard(self, dashboard_item):
        widget_map = {
            "indicatorWidget": arcgis.apps.dashboard.Indicator,
            "gaugeWidget": arcgis.apps.dashboard.Gauge,
            "pieChartWidget": arcgis.apps.dashboard.PieChart,
            "serialChartWidget": arcgis.apps.dashboard.SerialChart,
            "detailsWidget": arcgis.apps.dashboard.Details,
            "richTextWidget": arcgis.apps.dashboard.RichText,
            "listWidget": arcgis.apps.dashboard.List,
            "embeddedContentWidget": arcgis.apps.dashboard.EmbeddedContent,
            "legendWidget": arcgis.apps.dashboard.MapLegend,
        }
        item_json = dashboard_item.get_data()
        for widget_json in item_json["widgets"]:
            type = widget_json["type"]
            self._widgets.append(widget_map[type]._from_json(widget_json))

        return dashboard_item

    @staticmethod
    def _publish_random(widget):
        from arcgis.apps.dashboard import add_row

        gis = arcgis.env.active_gis
        import random
        import string

        letters = string.ascii_lowercase
        title = "".join(random.choice(letters) for i in range(10))
        summary = "".join(random.choice(letters) for i in range(10))
        db = Dashboard()

        if widget.type == "headerPanel":
            db.header = widget
        elif widget.type == "leftPanel":
            db.side_panel = widget
        elif widget.type == "legendWidget":
            map_widget_width = widget._map_widget.width
            widget._map_widget.width = 0.5
            db.layout = add_row([widget._map_widget, widget])
            widget._map_widget.width = map_widget_width
        else:
            if (
                hasattr(widget, "item")
                and getattr(widget.item, "type", None) == "mapWidget"
            ):
                db.layout = add_row([widget.item, widget])
            else:
                db.layout = add_row([widget])

        db = db.save(title, summary, gis=gis, overwrite=True)
        _created_dashboards.append((gis, db))
        url = f"{gis.url}/apps/opsdashboard/index.html#/{db.itemid}"

        return url


@atexit.register
def _delete_objects():
    for dashboard in _created_dashboards:
        gis = dashboard[0]
        db = dashboard[1]
        db.delete(force=True)



import os
import sys
import logging as _logging
import urllib.parse
import concurrent.futures
from types import MethodType
from functools import lru_cache
from arcgis.auth.tools import LazyLoader
from arcgis.gis import GIS
from arcgis import env

arcgis = LazyLoader("arcgis")
_mixins = LazyLoader("arcgis._impl.common._mixins")

from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoprocessing._tool import (
    _camelCase_to_underscore,
    _generate_param,
    _inspect_tool,
)

_log = _logging.getLogger(__name__)


###########################################################################
def _input_string_params(spec, name_type, name_param, num_spaces=20):
    """creates the input strings for the lambda"""
    src_code = ""
    optional_code = ""
    param_inputs = ""
    default_db = {}
    if len(spec) > 0:
        param_name, param_dval = spec[0]
        param_type = name_type[param_name]

        # src_code += _generate_param(name_param, param_dval, param_name, param_type)
        # param_inputs += f"{param_name}={param_name},"

        for param_name_dval in spec[:]:  # [ (param_name, param_dval) ]
            param_name, param_dval = param_name_dval
            param_type = name_type[param_name]
            if param_dval:
                optional_code += f"{param_name}=None,"
                default_db[param_name] = param_dval
            else:
                src_code += f"{param_name}"
            # src_code += _generate_param(name_param, param_dval, param_name, param_type)
            src_code += ","
            if src_code.find(",,") > -1:
                src_code = src_code.replace(",,", ",")
            param_inputs += f"{param_name}={param_name},"
        # src_code += ","
    src_code += ""
    if src_code.endswith(",") and optional_code:
        src_code += optional_code
    elif src_code.endswith(",") == False and optional_code:
        src_code += ","
        src_code += optional_code
    return src_code, param_inputs, default_db


###########################################################################
def _build_lambda(self, input_strings, param_inputs):
    """builds the lambda"""
    if input_strings.startswith(","):
        input_strings = input_strings[1:]

    return eval(
        f"lambda self, {input_strings}: self._run_tool({param_inputs})".replace(
            "\n", ""
        )
        .replace('''"""''', "'")
        .replace(":str", "")
    )


###########################################################################
class GPInfo:
    """
    Provides access to additional information about the GP Service.
    """

    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    @property
    def item_info(self) -> dict:
        """

        :return: the service's item information

        """
        url = f"{self._url}/iteminfo"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    @property
    def metadata(self) -> str:
        """

        :return: the service's metadata

        """
        url = f"{self._url}/metadata"

        return self._gis._con.get(url, {}, try_json=False)

    @property
    def thumbnail(self) -> str:
        """

        :return: the service's thumbnail

        """
        url = f"{self._url}/thumbnail"
        return self._gis._con.get(url, {}, try_json=False)


###########################################################################
class GPTask:
    """
    The GP Task resource represents a single task in a geoprocessing
    service published using ArcGIS Server. It provides basic information
    about the task including its name and display name. It also provides
    detailed information about the various input and output parameters
    exposed by the task.
    """

    _gis = None
    _url = None
    _parent = None
    _properties = None
    _choice_list = None
    _default_db = None

    def __init__(
        self,
        url: str,
        parent: "GPService",
        gis: GIS = None,
    ):
        self._parent = parent
        self._url = url
        if gis is None and env.active_gis is None:
            gis = GIS()
        elif gis is None and env.active_gis:
            gis = env.active_gis
        self._gis = gis
        name = self.properties["name"]
        uses_map_as_result = self._parent.properties.resultMapServerName != ""

        (
            helpstring,
            name_name,
            name_type,
            return_values,
            spec,
            name_param,
            choice_list_db_param,
        ) = _inspect_tool(self.properties, uses_map_as_result)
        param_db = {}
        for param_name_dval in spec:  # [ (param_name, param_dval) ]
            param_name, param_dval = param_name_dval
            param_type = name_type[param_name]
            gp_param_name = name_name[param_name]
            param_db[param_name] = (param_type.__name__, gp_param_name)
        return_values2 = []
        for retval in return_values:
            param_db[retval["name"]] = (retval["type"].__name__, retval["display_name"])
            return_values2.append(
                {
                    "name": retval["name"],
                    "display_name": retval["display_name"],
                    "type": retval["type"].__name__,
                }
            )
        self._return_values = return_values2
        input_string, param_string, self._default_db = _input_string_params(
            spec, name_type, name_param
        )
        self._param_db = param_db
        l = _build_lambda(
            self=self, input_strings=input_string, param_inputs=param_string
        )
        l.__doc__ = helpstring

        setattr(self, _camelCase_to_underscore(name), MethodType(l, self))

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    def _run_tool(self, **kwargs):
        """runs the tool"""
        future = True
        param_db = self._param_db
        return_values = self._return_values
        run_async = (
            self._parent.properties["executionType"] != "esriExecutionTypeSynchronous"
        )
        for key, value in kwargs.items():
            if key in self._default_db and value is None:
                kwargs[key] = self._default_db[key]

        return _execute_gp_tool(
            self._gis,
            os.path.basename(self._url),
            kwargs,
            param_db,
            return_values,
            run_async,
            os.path.dirname(self._url),
            future=future,
        )

    # ----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def properties(self) -> dict:
        """
        Returns the Service's Properties

        :return: Dictionary
        """
        if self._properties is None:
            params = {"f": "json"}
            self._properties = _mixins.PropertyMap(
                self._gis._con.get(self._url, params)
            )
        return self._properties

    # ----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def help_url(self) -> str:
        """
        Returns the URL to the documentation for the GP Task.

        :return: string
        """
        return self.properties.get("helpUrl", "")

    # ----------------------------------------------------------------------
    def default_value(self, name: str) -> object:
        """
        Gets a parameter's default value.


        :return: object

        """
        name = _camelCase_to_underscore(name=name)
        for param in self.properties["parameters"]:
            if _camelCase_to_underscore(param["name"]) == name:
                return param["defaultValue"]
        return None

    # ----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def name(self) -> str:
        """
        Name of the geoprocessing tasks

        :return: string

        """
        return _camelCase_to_underscore(self.properties["name"])

    # ----------------------------------------------------------------------
    @property
    @lru_cache(maxsize=10)
    def choice_list(self) -> dict:
        """
        Returns a Map of Parameters with Choice Lists.

        :return: Dictionary

        """
        if self._choice_list is None:
            self._choice_list = {}

            for param in self.properties["parameters"]:
                name = _camelCase_to_underscore(name=param["name"])
                cl = param.get("choiceList", None)
                if cl:
                    self._choice_list[name] = cl
        return self._choice_list


###########################################################################
class GPService:
    """
    A geoprocessing service can contain one or more tools that use input
    data from a client application, process it, and return output in the
    form of features, maps, reports, files, or services. These tools are
    first authored and run in ArcGIS Pro or ArcGIS Desktop, typically as
    a custom model or script tools, before being shared to an ArcGIS
    Server.
    """

    _gis = None
    _url = None
    _info = None
    _tasks = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS = None):
        self._url = url
        if gis is None and env.active_gis is None:
            gis = GIS()
        elif gis is None and env.active_gis:
            gis = env.active_gis
        self._gis = gis

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        Returns the Service's Properties

        :return: Dictionary
        """
        if self._properties is None:
            params = {"f": "json"}
            self._properties = _mixins.PropertyMap(
                self._gis._con.get(self._url, params)
            )
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def tasks(self) -> list:
        """returns the :class:`GP Tasks <arcgis.geoprocessing.GPTask>`"""
        if self._tasks is None:
            self._tasks = [
                GPTask(
                    url=self._url + urllib.parse.quote(f"/{task}"),
                    gis=self._gis,
                    parent=self,
                )
                for task in self.properties["tasks"]
            ]
        return self._tasks

    # ----------------------------------------------------------------------
    def refresh(self):
        """
        Reloads the Service Information
        """
        self._tasks = None
        self._properties = None

    # ----------------------------------------------------------------------
    @property
    def info(self) -> GPInfo:
        """
        :return:
            :class:`~arcgis.geoprocessing.GPInfo`
        """
        if self._info is None:
            url = f"{self._url}/info"
            self._info = GPInfo(url, self._gis)
        return self._info
